"Tests using the example extensions we registered in setup.py."

import kontakt
import kontakt.examples.extension

from helpers import EXTENSION_TYPES, EXTENSION_NAMESPACE


def test_list_extensions():
    actual = sorted(kontakt.list_extensions(EXTENSION_NAMESPACE))
    expected = sorted(EXTENSION_TYPES)
    assert actual == expected



import importlib.abc
import importlib.resources
import inspect
import textwrap
from abc import ABC, abstractmethod
from typing import Mapping

import asq
import stevedore
import stevedore.exception


class Extension(ABC):
    """Base class for new extension types.

    Args:
        name: The name of the extension, i.e. as assigned in entry-points.
    """

    def __init__(self, name):
        self._name = name

    @classmethod
    def kind(cls) -> str:
        "The kind of the extension."
        return cls._kind()

    @classmethod
    @abstractmethod
    def _kind(cls) -> str:
        "Override in subclasses to specify the kind of the extension."
        raise NotImplementedError

    @property
    def name(self):
        "The name of the extension instance."
        return self._name

    @classmethod
    def dirpath(cls):
        """The directory path to the extension package."""
        package_name = inspect.getmodule(cls).__package__
        return importlib.resources.files(package_name)

    @classmethod
    def version(cls):
        "The version of the extension."
        return "1.0.0"  # We allow extensions to have a distinct version but don't exploit this yet

    @classmethod
    def describe(cls) -> str:
        """A description of the extension.

        By default, this is the docstring of the extension class if the docstring is defined. If
        the docstring is not defined, the default is the name of the class.

        Override it in the extension if you want something different.
        """

        if cls.__doc__ is None:
            description = cls.__name__
        else:
            description = cls.__doc__

        return _strip_lines(textwrap.dedent(description))

    @classmethod
    def new(cls, kind: str):
        """Decorator for creating new Extension subclasses.

        Use it like this:

            @Extension.new('renderer')
            class RendererExtension:
                @abstractmethod
                def render(self):
                    return NotImplementedError

            assert issubclass(RendererExtension, Extension)
            assert RendererExtension.kind() == 'renderer'

        Args:
            kind: The 'kind' string of the extension. This is just descriptive.
        """

        def decorator(decorated_cls):
            class Ext(decorated_cls, cls):
                @classmethod
                def _kind(cls):
                    return kind

            return Ext

        return decorator


class ExtensionError(Exception):
    "Base class for extension-specific exceptions."
    pass


def create_extension(namespace: str, name: str, exception_type: Exception, *args, **kwargs):
    """Create a new instance of an extension.

    Args:
        namespace (str): The namespace of the extension to create.
        name (str): The name of the extension to create.
        exception_type (Exception): Exception to raise when there are errors creating the exception.

    Raises:
        exception_type: If there is an error creating the extension.

    Returns:
        Any: Whatever type of object the extension provides. Typically this will be an Extension instance.
    """
    normal_name = _normalize_name(name)
    try:
        manager = stevedore.driver.DriverManager(
            namespace=namespace,
            name=normal_name,
            invoke_on_load=True,
            invoke_args=args,
            invoke_kwds={**kwargs, "name": normal_name},
        )
    except stevedore.exception.NoMatches as no_matches:
        names = list_extensions(namespace)
        name_list = ", ".join(names)
        raise exception_type(f"No {namespace} matching {name !r}. Available {namespace}s: {name_list}") from no_matches
    driver = manager.driver
    return driver


def describe_extension(namespace, name, exception_type=None) -> str:
    """Get the description of an extension.

    Args:
        namespace (str): The namespace of the extension to create.
        name (str): The name of the extension to create.
        exception_type (Exception): Exception to raise when there are errors creating the exception.

    Raises:
        exception_type: If there is an error creating the extension.

    Returns:
        str: Description of the plugin.
    """
    exception_type = exception_type or ExtensionError
    normal_name = _normalize_name(name)
    try:
        manager = stevedore.driver.DriverManager(
            namespace=namespace,
            name=normal_name,
            invoke_on_load=False,
        )
    except stevedore.exception.NoMatches as no_matches:
        names = list_extensions(namespace)
        name_list = ", ".join(names)
        raise exception_type(f"No {namespace} matching {name !r}. Available {namespace}s: {name_list}") from no_matches
    driver = manager.driver
    description = driver.describe()
    return description


def list_extensions(namespace):
    """List the names of the extensions available in a given namespace."""
    extensions = stevedore.ExtensionManager(
        namespace=namespace,
        invoke_on_load=False,
    )
    return extensions.names()


def list_dirpaths(namespace) -> Mapping[str, importlib.abc.Traversable]:
    """A mapping of extension names to extension package paths."""
    extensions = stevedore.ExtensionManager(
        namespace=namespace,
        invoke_on_load=False,
    )
    return {name: _extension_dirpath(ext) for name, ext in extensions.items()}


def _extension_dirpath(ext: stevedore.extension.Extension) -> importlib.abc.Traversable:
    """Get the directory path to an extension package.

    Args:
        ext: A stevedore.extension.Extension instance.

    Returns:
        A absolute Path to the package containing the extension.
    """
    package_name = ext.module_name.split(".")[0]
    resource = importlib.resources.files(package_name)
    return resource.resolve()


def _normalize_name(name):
    """Normalise a name (such as a master name) by converting hyphens to underscores."""
    return name.replace("-", "_")


def _strip_lines(text):
    """Remove leading and trailing blank lines."""

    def is_blank(line):
        return line.isspace() or not line

    return "\n".join(asq.query(text.splitlines()).skip_while(is_blank).reverse().skip_while(is_blank).reverse())



"Tests using the example extensions we registered in setup.py."

import kontakt
import kontakt.examples.extension
import pytest

from helpers import EXTENSION_NAMESPACE


def test_raises_correct_exception_when_no_extension_exists():
    with pytest.raises(kontakt.examples.extension.ExampleExtensionError):
        kontakt.create_extension(
            namespace=EXTENSION_NAMESPACE,
            name="no-such-extension",
            exception_type=kontakt.examples.extension.ExampleExtensionError,
        )


def test_creates_correct_type_of_extension(extension_type):
    name, extension_type = extension_type

    extension = kontakt.create_extension(
        EXTENSION_NAMESPACE,
        name,
        kontakt.examples.extension.ExampleExtensionError,
        "flavor",
        42,
    )

    assert isinstance(extension, extension_type)


def test_forwards_positional_arguments(extension_type):
    name, extension_type = extension_type
    flavor = f"{name} flavored"
    size = 42
    extension = kontakt.create_extension(
        EXTENSION_NAMESPACE,
        name,
        kontakt.examples.extension.ExampleExtensionError,
        flavor,
        size,
    )

    assert extension.flavor == flavor
    assert extension.size == size


def test_forwards_keyword_arguments(extension_type):
    name, extension_type = extension_type
    flavor = f"{name} flavored"
    size = 42
    extension = kontakt.create_extension(
        EXTENSION_NAMESPACE,
        name,
        kontakt.examples.extension.ExampleExtensionError,
        flavor=flavor,
        size=size,
    )

    assert extension.flavor == flavor
    assert extension.size == size


def test_forwards_mixed_positional_and_keyword_arguments(extension_type):
    name, extension_type = extension_type
    flavor = f"{name} flavored"
    size = 42
    extension = kontakt.create_extension(
        EXTENSION_NAMESPACE,
        name,
        kontakt.examples.extension.ExampleExtensionError,
        flavor,
        size=size,
    )

    assert extension.flavor == flavor
    assert extension.size == size


def test_raises_type_error_when_passed_unknown_keyword_argument(extension_type):
    name, extension_type = extension_type
    with pytest.raises(TypeError):
        kontakt.create_extension(
            EXTENSION_NAMESPACE,
            name,
            kontakt.examples.extension.ExampleExtensionError,
            "raspberry",
            azimuth=123,
        )


def test_raises_type_error_when_passed_too_many_positional_arguments(extension_type):
    name, extension_type = extension_type
    with pytest.raises(TypeError):
        kontakt.create_extension(
            EXTENSION_NAMESPACE,
            name,
            kontakt.examples.extension.ExampleExtensionError,
            "raspberry",
            123,
            "kromulent",
        )


def test_raises_type_error_when_passed_too_few_positional_arguments(extension_type):
    name, extension_type = extension_type
    with pytest.raises(TypeError):
        kontakt.create_extension(
            EXTENSION_NAMESPACE,
            name,
            kontakt.examples.extension.ExampleExtensionError,
            "raspberry",
        )



"""Example of how to create a ConfigurableExtension."""

from dataclasses import dataclass
from kontakt.config_options import ConfigurableExtension, Option
from kontakt import ExtensionError

KIND = "renderer"


@ConfigurableExtension.new(KIND)
class RendererExtension(ConfigurableExtension):
    pass


class RendererExtensionError(ExtensionError):
    pass


class PrinterExtension(RendererExtension):
    def config_options(self):
        return (Option(("font", "size"), "Font size", 42), Option(("font", "family"), "Font family", "serif"))

    def construct(self, extension_config, *args, **kwargs):
        return Printer(
            font_size=extension_config["font"]["size"], font_family=extension_config["font"]["family"], *args, **kwargs
        )


@dataclass
class Printer:
    font_size: int
    font_family: str
    model: str



import kontakt
from kontakt.config_options import Option, build_default_config
from kontakt.examples.configurable_extension import (Printer,
                                                     RendererExtensionError)

from helpers import CONFIGURABLE_EXTENSION_NAMESPACE


def test_config_options():
    extension = kontakt.create_extension(
        CONFIGURABLE_EXTENSION_NAMESPACE,
        "printer",
        RendererExtensionError,
    )

    actual = list(extension.config_options())
    expected = [Option(("font", "size"), "Font size", 42), Option(("font", "family"), "Font family", "serif")]
    assert actual == expected


def test_construct_object():
    extension = kontakt.create_extension(
        CONFIGURABLE_EXTENSION_NAMESPACE,
        "printer",
        RendererExtensionError,
    )

    config = build_default_config(extension.config_options())

    actual = extension.construct(config, model="PrinterCo")
    expected = Printer(font_size=42, font_family="serif", model="PrinterCo")
    assert actual == expected



"""Support for defining configurable options and building default configs from them.

There are two related ideas here. The first is of a *configurable options*, a description of
a single option in a *configuration*. A configuration, then, is just a tree - modeled as
dicts-of-dicts - where the leaf values are the options. So an option describes a path through
the tree to a leaf, a description of the option, and a default value for the leaf.
"""

from abc import abstractmethod
from dataclasses import dataclass
import copy
from typing import Any, Iterable, Mapping, Tuple

from dendrodict import DeepDict

from . import Extension


@dataclass
class Option:
    """Description of a single configurable option.

    Fields:
        path: Path to the option in the config.
        description: Description of the option.
        default: Default value of the option.
    """

    path: Tuple[str]
    description: str
    default: Any

    def __post_init__(self):
        if len(self.path) < 1:
            raise ValueError("Path must have at least one element")

    @property
    def name(self):
        "The name of the option."
        return self.path[-1]


class ConfigurableExtension(Extension):
    @abstractmethod
    def construct(self, extension_config, *args, **kwargs):
        """Construct the object that the extension provides.

        This method can use config values in `extension_config` as well as
        values from *args and **kwargs to build the object.

        Args:
            extension_config (dict): Config dict for the extension. It should
                have value for each of the extension's ``Option``\s.

        Returns:
            Any: The object provided by this extension.
        """
        return NotImplementedError

    def config_options(self) -> Iterable[Option]:
        """The configuration options used by this extension.

        Returns:
            Iterable[Option]: An iterable of ``Option``\s describing the
                configuration options used by the extension.
        """
        return ()


def build_default_config(options) -> dict:
    """Build a default config from an iterable of options."""
    return DeepDict({option.path: option.default for option in options}).to_dict()


def merge_configs(dest, src):
    """Merge two config dicts.

    Merging can't happen if the dictionaries are incompatible. This happens when the same path in `src` exists in `dest`
    and one points to a `dict` while another points to a non-`dict`.

    Returns: A new `dict` with the contents of `src` merged into `dest`.

    Raises:
        ValueError: If the two dicts are incompatible.
    """
    dest = copy.deepcopy(dest)

    for src_name, src_val in src.items():
        if isinstance(src_val, Mapping):
            dest_val = dest.get(src_name, {})
            if not isinstance(dest_val, Mapping):
                raise ValueError("Incompatible config structures")

            dest[src_name] = merge_configs(dest_val, src_val)
        else:
            try:
                if isinstance(dest[src_name], Mapping):
                    raise ValueError("Incompatible config structures")
            except KeyError:
                pass
            dest[src_name] = src_val

    return dest



from pathlib import Path

import kontakt

from helpers import EXTENSION_TYPES, EXTENSION_NAMESPACE


def test_list_dirpaths():
    actual = kontakt.list_dirpaths(EXTENSION_NAMESPACE)
    expected = {name: Path(kontakt.__file__).parent.resolve() for name in EXTENSION_TYPES}
    assert actual == expected



"""Examples of how to use the kontakt extension system.

These serve both as a concrete example of how to use kontakt
as well as for use in the tests.
"""

import kontakt


# First we define a new kind of extension, 'example-extension', of which we'll create different implementations below.
@kontakt.Extension.new('example-extension')
class ExampleExtension:
    def __init__(self, flavor, size, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.flavor = flavor
        self.size = size


class ExampleExtensionError(kontakt.ExtensionError):
    pass


# Here's the first implementation of ExampleExtension. This would be registered as an entry point in setup.py something like this:
#
#    entry_points = {
#        "my_app.examples": [ "blue = my_app.examples.blue.Blue", . . .]
#    }
class Blue(ExampleExtension):
    pass

# This second one would be registered something like this: "green = my_app.examples.red.Red"
class Red(ExampleExtension):
    "Red extension"
    pass


class Green(ExampleExtension):
    @classmethod
    def describe(cls):
        return "description of the Green extension"
        
# Then you'd probably have a function like this for constructing the ExampleExtension subclasses

def get_example_extension(name):
    return kontakt.create_extension(
        namespace='my_app.examples',
        name=name, # e.g. blue, red, etc...whatever names you used in entry_points.my_app.examples
        exception_type=ExampleExtensionError
    )



from pathlib import Path

import kontakt
from kontakt.examples.extension import ExampleExtension, ExampleExtensionError

from helpers import EXTENSION_NAMESPACE

# TODO: Many of these tests would work for configurable extensions as well. Should we include them?


def test_kind_attribute(extension_type):
    name, extension_type = extension_type

    extension = kontakt.create_extension(
        EXTENSION_NAMESPACE,
        name,
        ExampleExtensionError,
        "flavor",
        42,
    )

    assert extension.kind() == ExampleExtension.kind()


def test_name_attribute(extension_type):
    name, extension_type = extension_type

    extension = kontakt.create_extension(
        EXTENSION_NAMESPACE,
        name,
        ExampleExtensionError,
        "flavor",
        42,
    )

    assert extension.name == name


def test_dirpath_attribute(extension_type):
    name, extension_type = extension_type

    assert Path(extension_type.dirpath()).resolve() == (Path(kontakt.__file__).parent / "examples").resolve()


def test_version_property(extension_type):
    name, extension_type = extension_type

    extension = kontakt.create_extension(
        EXTENSION_NAMESPACE,
        name,
        ExampleExtensionError,
        "flavor",
        42,
    )

    assert extension.version() == "1.0.0"


def test_description_defaults_to_class_name_without_docstring():
    extension = kontakt.create_extension(
        EXTENSION_NAMESPACE,
        "blue",
        ExampleExtensionError,
        "blueberry",
        42,
    )

    actual = extension.describe()

    expected = kontakt.examples.extension.Blue.__name__

    assert actual == expected


def test_description_defaults_to_docstring_if_available():
    extension = kontakt.create_extension(
        EXTENSION_NAMESPACE,
        "red",
        ExampleExtensionError,
        "raspberry",
        42,
    )

    actual = extension.describe()

    expected = kontakt.examples.extension.Red.__doc__

    assert actual == expected


def test_extensions_can_override_description():
    extension = kontakt.create_extension(
        EXTENSION_NAMESPACE,
        "green",
        ExampleExtensionError,
        "apple",
        42,
    )

    actual = extension.describe()

    expected = kontakt.examples.extension.Green("apple", 42, "green").describe()

    assert actual == expected



"Tests using the example extensions we registered in setup.py."

import kontakt
import kontakt.examples.extension
import pytest

from helpers import EXTENSION_NAMESPACE


def test_raises_correct_exception_when_no_extension_exists():
    with pytest.raises(kontakt.examples.extension.ExampleExtensionError):
        kontakt.describe_extension(
            namespace=EXTENSION_NAMESPACE,
            name="no-such-extension",
            exception_type=kontakt.examples.extension.ExampleExtensionError,
        )


def test_description_defaults_to_class_name_without_docstring():
    actual = kontakt.describe_extension(
        EXTENSION_NAMESPACE,
        "blue",
        kontakt.examples.extension.ExampleExtensionError,
    )

    expected = kontakt.examples.extension.Blue.__name__

    assert actual == expected


def test_description_defaults_to_docstring_if_available():
    actual = kontakt.describe_extension(
        EXTENSION_NAMESPACE,
        "red",
        kontakt.examples.extension.ExampleExtensionError,
    )

    expected = kontakt.examples.extension.Red.__doc__

    assert actual == expected


def test_extensions_can_override_description():
    actual = kontakt.describe_extension(
        EXTENSION_NAMESPACE,
        "green",
        kontakt.examples.extension.ExampleExtensionError,
    )

    expected = kontakt.examples.extension.Green.describe()

    assert actual == expected
